/**
 * SQL-DK
 * Copyright © 2013 František Kučera (frantovo.cz)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package info.globalcode.sql.dk;

import info.globalcode.sql.dk.configuration.Configuration;
import info.globalcode.sql.dk.configuration.ConfigurationException;
import info.globalcode.sql.dk.configuration.ConfigurationProvider;
import info.globalcode.sql.dk.configuration.DatabaseDefinition;
import info.globalcode.sql.dk.configuration.FormatterDefinition;
import info.globalcode.sql.dk.configuration.Properties;
import info.globalcode.sql.dk.configuration.Property;
import info.globalcode.sql.dk.formatting.ColumnsHeader;
import info.globalcode.sql.dk.formatting.FakeSqlArray;
import info.globalcode.sql.dk.formatting.Formatter;
import info.globalcode.sql.dk.formatting.FormatterContext;
import info.globalcode.sql.dk.formatting.FormatterException;
import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.sql.Array;
import java.sql.Driver;
import java.sql.DriverPropertyInfo;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.List;
import java.util.ServiceLoader;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.sql.rowset.RowSetMetaDataImpl;

/**
 * Displays info like help, version etc.
 *
 * @author Ing. František Kučera (frantovo.cz)
 */
public class InfoLister {

	private static final Logger log = Logger.getLogger(InfoLister.class.getName());
	/**
	 * Fake database name for output formatting
	 */
	public static final String CONFIG_DB_NAME = "sqldk_configuration";
	private PrintStream out;
	private ConfigurationProvider configurationProvider;
	private CLIOptions options;
	private Formatter formatter;

	public InfoLister(PrintStream out, ConfigurationProvider configurationProvider, CLIOptions options) {
		this.out = out;
		this.configurationProvider = configurationProvider;
		this.options = options;
	}

	public void showInfo() throws ConfigurationException, FormatterException {
		EnumSet<InfoType> commands = options.getShowInfo();

		boolean formattinNeeded = false;

		for (InfoType infoType : commands) {
			switch (infoType) {
				case CONNECTION:
				case JDBC_DRIVERS:
				case JDBC_PROPERTIES:
				case DATABASES:
				case FORMATTERS:
				case TYPES:
					formattinNeeded = true;
					break;
			}
		}

		if (formattinNeeded) {
			try (Formatter f = getFormatter()) {
				formatter = f;
				formatter.writeStartBatch();
				DatabaseDefinition dd = new DatabaseDefinition();
				dd.setName(CONFIG_DB_NAME);
				formatter.writeStartDatabase(dd);
				showInfos(commands);
				formatter.writeEndDatabase();
				formatter.writeEndBatch();
				formatter.close();
			}
		} else {
			showInfos(commands);
		}
	}

	private void showInfos(EnumSet<InfoType> commands) throws ConfigurationException, FormatterException {
		for (InfoType infoType : commands) {
			infoType.showInfo(this);
		}
	}

	private void listFormatters() throws ConfigurationException, FormatterException {
		ColumnsHeader header = constructHeader(
				new HeaderField("name", SQLType.VARCHAR),
				new HeaderField("built_in", SQLType.BOOLEAN),
				new HeaderField("default", SQLType.BOOLEAN),
				new HeaderField("class_name", SQLType.VARCHAR),
				new HeaderField("valid", SQLType.BOOLEAN));
		List<Object[]> data = new ArrayList<>();

		String defaultFormatter = configurationProvider.getConfiguration().getDefaultFormatter();
		defaultFormatter = defaultFormatter == null ? Configuration.DEFAULT_FORMATTER : defaultFormatter;

		for (FormatterDefinition fd : configurationProvider.getConfiguration().getBuildInFormatters()) {
			data.add(new Object[]{fd.getName(), true, defaultFormatter.equals(fd.getName()), fd.getClassName(), isInstantiable(fd)});
		}

		for (FormatterDefinition fd : configurationProvider.getConfiguration().getFormatters()) {
			data.add(new Object[]{fd.getName(), false, defaultFormatter.equals(fd.getName()), fd.getClassName(), isInstantiable(fd)});
		}

		printTable(formatter, header, data, "-- configured and built-in output formatters", null);
	}

	private boolean isInstantiable(FormatterDefinition fd) {
		try {
			try (ByteArrayOutputStream testStream = new ByteArrayOutputStream()) {
				fd.getInstance(new FormatterContext(testStream, new Properties(0)));
				return true;
			}
		} catch (Exception e) {
			log.log(Level.SEVERE, "Unable to create an instance of formatter: " + fd.getName(), e);
			return false;
		}
	}

	public void listTypes() throws FormatterException, ConfigurationException {
		ColumnsHeader header = constructHeader(new HeaderField("name", SQLType.VARCHAR), new HeaderField("code", SQLType.INTEGER));
		List<Object[]> data = new ArrayList<>();
		for (SQLType sqlType : SQLType.values()) {
			data.add(new Object[]{sqlType.name(), sqlType.getCode()});
		}
		printTable(formatter, header, data, "-- data types", null);
		log.log(Level.INFO, "Type names in --types option are case insensitive");
	}

	public void listDatabases() throws ConfigurationException, FormatterException {
		ColumnsHeader header = constructHeader(
				new HeaderField("database_name", SQLType.VARCHAR),
				new HeaderField("user_name", SQLType.VARCHAR),
				new HeaderField("database_url", SQLType.VARCHAR));
		List<Object[]> data = new ArrayList<>();

		final List<DatabaseDefinition> configuredDatabases = configurationProvider.getConfiguration().getDatabases();
		if (configuredDatabases.isEmpty()) {
			log.log(Level.WARNING, "No databases are configured.");
		} else {
			for (DatabaseDefinition dd : configuredDatabases) {
				data.add(new Object[]{dd.getName(), dd.getUserName(), dd.getUrl()});
			}
		}

		printTable(formatter, header, data, "-- configured databases", null);
	}

	public void listJdbcDrivers() throws FormatterException, ConfigurationException {
		ColumnsHeader header = constructHeader(
				new HeaderField("class", SQLType.VARCHAR),
				new HeaderField("version", SQLType.VARCHAR),
				new HeaderField("major", SQLType.INTEGER),
				new HeaderField("minor", SQLType.INTEGER),
				new HeaderField("jdbc_compliant", SQLType.BOOLEAN));
		List<Object[]> data = new ArrayList<>();

		final ServiceLoader<Driver> drivers = ServiceLoader.load(Driver.class);
		for (Driver d : drivers) {
			data.add(new Object[]{
				d.getClass().getName(),
				d.getMajorVersion() + "." + d.getMinorVersion(),
				d.getMajorVersion(),
				d.getMinorVersion(),
				d.jdbcCompliant()
			});
		}

		printTable(formatter, header, data, "-- discovered JDBC drivers (available on the CLASSPATH)", null);
	}

	public void listJdbcProperties() throws FormatterException, ConfigurationException {
		for (String dbName : options.getDatabaseNamesToListProperties()) {
			ColumnsHeader header = constructHeader(
					new HeaderField("property_name", SQLType.VARCHAR),
					new HeaderField("required", SQLType.BOOLEAN),
					new HeaderField("choices", SQLType.ARRAY),
					new HeaderField("configured_value", SQLType.VARCHAR),
					new HeaderField("description", SQLType.VARCHAR));
			List<Object[]> data = new ArrayList<>();

			DatabaseDefinition dd = configurationProvider.getConfiguration().getDatabase(dbName);

			Driver driver = findDriver(dd);

			if (driver == null) {
				log.log(Level.WARNING, "No JDBC driver was found for DB: {0} with URL: {1}", new Object[]{dd.getName(), dd.getUrl()});
			} else {
				log.log(Level.INFO, "For DB: {0} was found JDBC driver: {1}", new Object[]{dd.getName(), driver.getClass().getName()});

				try {
					DriverPropertyInfo[] propertyInfos = driver.getPropertyInfo(dd.getUrl(), dd.getProperties().getJavaProperties());

					Set<String> standardProperties = new HashSet<>();

					for (DriverPropertyInfo pi : propertyInfos) {
						Array choices = new FakeSqlArray(pi.choices, SQLType.VARCHAR);
						data.add(new Object[]{
							pi.name,
							pi.required,
							choices.getArray() == null ? "" : choices,
							pi.value == null ? "" : pi.value,
							pi.description
						});
						standardProperties.add(pi.name);
					}

					for (Property p : dd.getProperties()) {
						if (!standardProperties.contains(p.getName())) {
							data.add(new Object[]{
								p.getName(),
								"",
								"",
								p.getValue(),
								""
							});
							log.log(Level.WARNING, "Your configuration contains property „{0}“ not declared by the JDBC driver.", p.getName());
						}
					}

				} catch (SQLException e) {
					log.log(Level.WARNING, "Error during getting property infos.", e);
				}

				List<Parameter> parameters = new ArrayList<>();
				parameters.add(new NamedParameter("databgase", dbName, SQLType.VARCHAR));
				parameters.add(new NamedParameter("driver_class", driver.getClass().getName(), SQLType.VARCHAR));
				parameters.add(new NamedParameter("driver_major_version", driver.getMajorVersion(), SQLType.INTEGER));
				parameters.add(new NamedParameter("driver_minor_version", driver.getMinorVersion(), SQLType.INTEGER));

				printTable(formatter, header, data, "-- configured and configurable JDBC driver properties", parameters);
			}
		}

	}

	private Driver findDriver(DatabaseDefinition dd) {
		final ServiceLoader<Driver> drivers = ServiceLoader.load(Driver.class);
		for (Driver d : drivers) {
			try {
				if (d.acceptsURL(dd.getUrl())) {
					return d;
				}
			} catch (SQLException e) {
				log.log(Level.WARNING, "Error during finding JDBC driver for: " + dd.getName(), e);
			}
		}
		return null;
	}

	public void testConnection() throws FormatterException, ConfigurationException {
		ColumnsHeader header = constructHeader(
				new HeaderField("database_name", SQLType.VARCHAR),
				new HeaderField("configured", SQLType.BOOLEAN),
				new HeaderField("connected", SQLType.BOOLEAN));
		List<Object[]> data = new ArrayList<>();

		for (String dbName : options.getDatabaseNamesToTest()) {
			data.add(testConnection(dbName));
		}

		printTable(formatter, header, data, "-- database configuration and connectivity test", null);
	}

	public Object[] testConnection(String dbName) {
		log.log(Level.FINE, "Testing connection to database: {0}", dbName);

		boolean succesfullyConnected = false;
		boolean succesfullyConfigured = false;

		try {
			DatabaseDefinition dd = configurationProvider.getConfiguration().getDatabase(dbName);
			log.log(Level.FINE, "Database definition was loaded from configuration");
			succesfullyConfigured = true;
			try (DatabaseConnection dc = dd.connect(options.getDatabaseProperties())) {
				succesfullyConnected = dc.test();
			}
			log.log(Level.FINE, "Database connection test was successful");
		} catch (ConfigurationException | SQLException e) {
			log.log(Level.SEVERE, "Error during testing connection", e);
		}

		return new Object[]{dbName, succesfullyConfigured, succesfullyConnected};
	}

	public void printResource(String fileName) {
		try (BufferedReader reader = new BufferedReader(new InputStreamReader(getClass().getClassLoader().getResourceAsStream(fileName)))) {
			while (true) {
				String line = reader.readLine();
				if (line == null) {
					break;
				} else {
					println(line);
				}
			}
		} catch (Exception e) {
			log.log(Level.SEVERE, "Unable to print this info. Please see our website for it: " + Constants.WEBSITE, e);
		}
	}

	private void println(String line) {
		out.println(line);
	}

	private void printTable(Formatter formatter, ColumnsHeader header, List<Object[]> data, String sql, List<Parameter> parameters) throws ConfigurationException, FormatterException {
		formatter.writeStartStatement();

		if (sql != null) {
			formatter.writeQuery(sql);
			if (parameters != null) {
				formatter.writeParameters(parameters);
			}
		}

		formatter.writeStartResultSet(header);

		for (Object[] row : data) {
			formatter.writeStartRow();
			for (Object cell : row) {
				formatter.writeColumnValue(cell);
			}
			formatter.writeEndRow();
		}

		formatter.writeEndResultSet();
		formatter.writeEndStatement();
	}

	private Formatter getFormatter() throws ConfigurationException, FormatterException {
		String formatterName = options.getFormatterName();
		formatterName = formatterName == null ? Configuration.DEFAULT_FORMATTER_PREFETCHING : formatterName;
		FormatterDefinition fd = configurationProvider.getConfiguration().getFormatter(formatterName);
		FormatterContext context = new FormatterContext(out, options.getFormatterProperties());
		return fd.getInstance(context);
	}

	private ColumnsHeader constructHeader(HeaderField... fields) throws FormatterException {
		try {
			RowSetMetaDataImpl metaData = new RowSetMetaDataImpl();
			metaData.setColumnCount(fields.length);

			for (int i = 0; i < fields.length; i++) {
				HeaderField hf = fields[i];
				int sqlIndex = i + 1;
				metaData.setColumnName(sqlIndex, hf.name);
				metaData.setColumnLabel(sqlIndex, hf.name);
				metaData.setColumnType(sqlIndex, hf.type.getCode());
				metaData.setColumnTypeName(sqlIndex, hf.type.name());
			}

			return new ColumnsHeader(metaData);
		} catch (SQLException e) {
			throw new FormatterException("Error while constructing table headers", e);
		}
	}

	private static class HeaderField {

		String name;
		SQLType type;

		public HeaderField(String name, SQLType type) {
			this.name = name;
			this.type = type;
		}
	}

	public enum InfoType {

		HELP {
			@Override
			public void showInfo(InfoLister infoLister) {
				infoLister.printResource(Constants.HELP_FILE);
			}
		},
		VERSION {
			@Override
			public void showInfo(InfoLister infoLister) {
				infoLister.printResource(Constants.VERSION_FILE);
			}
		},
		LICENSE {
			@Override
			public void showInfo(InfoLister infoLister) {
				infoLister.printResource(Constants.LICENSE_FILE);
			}
		},
		FORMATTERS {
			@Override
			public void showInfo(InfoLister infoLister) throws FormatterException, ConfigurationException {
				infoLister.listFormatters();
			}
		},
		TYPES {
			@Override
			public void showInfo(InfoLister infoLister) throws FormatterException, ConfigurationException {
				infoLister.listTypes();
			}
		},
		JDBC_DRIVERS {
			@Override
			public void showInfo(InfoLister infoLister) throws ConfigurationException, FormatterException {
				infoLister.listJdbcDrivers();
			}
		},
		JDBC_PROPERTIES {
			@Override
			public void showInfo(InfoLister infoLister) throws ConfigurationException, FormatterException {
				infoLister.listJdbcProperties();
			}
		},
		DATABASES {
			@Override
			public void showInfo(InfoLister infoLister) throws FormatterException, ConfigurationException {
				infoLister.listDatabases();
			}
		},
		CONNECTION {
			@Override
			public void showInfo(InfoLister infoLister) throws FormatterException, ConfigurationException {
				infoLister.testConnection();
			}
		};

		public abstract void showInfo(InfoLister infoLister) throws ConfigurationException, FormatterException;
	}
}
