/**
 * SQL-DK
 * Copyright © 2014 František Kučera (frantovo.cz)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package info.globalcode.sql.dk.formatting;

import info.globalcode.sql.dk.Constants;
import info.globalcode.sql.dk.NamedParameter;
import info.globalcode.sql.dk.Parameter;
import info.globalcode.sql.dk.Xmlns;
import info.globalcode.sql.dk.configuration.DatabaseDefinition;
import info.globalcode.sql.dk.configuration.Properties;
import info.globalcode.sql.dk.configuration.Property;
import static info.globalcode.sql.dk.formatting.AbstractXmlFormatter.qname;
import java.sql.Array;
import java.sql.SQLException;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Scanner;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.namespace.QName;

/**
 * Prints result sets and parameters as tables, SQL as preformatted and updates counts as
 * paragraphs. You can pick XHTML fragments (usually tabular data) and use it on your website or use
 * whole output as preview or report.
 *
 * @author Ing. František Kučera (frantovo.cz)
 */
public class XhtmlFormatter extends AbstractXmlFormatter {

	private static final Logger log = Logger.getLogger(XhtmlFormatter.class.getName());
	public static final String NAME = "xhtml"; // bash-completion:formatter
	private static final String DOCTYPE = "html PUBLIC \"-//W3C//DTD XHTML 1.1 plus MathML 2.0 plus SVG 1.1//EN\" \"http://www.w3.org/2002/04/xhtml-math-svg/xhtml-math-svg.dtd\"";
	private static final String CSS_FILE = "info/globalcode/sql/dk/formatter/XhtmlFormatter.css";
	private int statementCounter = 0;
	private int resultSetCounter = 0;
	private int updatesResultCounter = 0;

	public XhtmlFormatter(FormatterContext formatterContext) {
		super(addDefaults(formatterContext));
	}

	/**
	 * Do not indent text – preserve whitespace for pre elements
	 */
	private static FormatterContext addDefaults(FormatterContext formatterContext) {
		Properties defaults = new Properties(1);
		defaults.add(new Property(PROPERTY_INDENT_TEXT, "false"));
		formatterContext.getProperties().setLastDefaults(defaults);
		return formatterContext;
	}

	@Override
	public void writeStartBatch() {
		super.writeStartBatch();
		printStartDocument();
		printDoctype(DOCTYPE);
		printStartElement(qname("html"), singleAttribute(qname("xmlns"), Xmlns.XHTML));

		printStartElement(qname("head"));
		printTextElement(qname("title"), null, Constants.PROGRAM_NAME + ": batch results");
		printCss();
		printEndElement();

		printStartElement(qname("body"));
	}

	private void printCss() {

		try (Scanner css = new Scanner(getClass().getClassLoader().getResourceAsStream(CSS_FILE))) {
			printStartElement(qname("style"), singleAttribute(qname("type"), "text/css"));
			while (css.hasNext()) {
				printText(css.nextLine(), true);
			}
			printEndElement();
		}
	}

	@Override
	public void writeEndBatch() {
		super.writeEndBatch();
		printEndElement();
		printEndElement();
		printEndDocument();
	}

	@Override
	public void writeStartDatabase(DatabaseDefinition databaseDefinition) {
		super.writeStartDatabase(databaseDefinition);
		printTextElement(qname("h1"), null, "Database: " + databaseDefinition.getName());

		printStartElement(qname("p"));
		printText("This is XHTML output of batch executed at: ", true);
		printText(new Date().toString(), true);
		printEndElement();
	}

	@Override
	public void writeQuery(String sql) {
		super.writeQuery(sql);
		printTextElement(qname("pre"), null, sql);
	}

	@Override
	public void writeParameters(List<? extends Parameter> parameters) {
		super.writeParameters(parameters);

		if (parameters == null || parameters.isEmpty()) {
			printTextElement(qname("p"), null, "(this query has no parameters)");
		} else {
			printTextElement(qname("h3"), null, "Parameters:");

			printStartElement(qname("table"));

			printStartElement(qname("thead"));
			printStartElement(qname("tr"));
			printTextElement(qname("td"), null, "id");
			printTextElement(qname("td"), null, "type");
			printTextElement(qname("td"), null, "value");
			printEndElement();
			printEndElement();

			printStartElement(qname("tbody"));
			for (int i = 0; i < parameters.size(); i++) {
				Parameter p = parameters.get(i);
				printStartElement(qname("tr"));
				String numberOrName;
				if (p instanceof NamedParameter) {
					numberOrName = ((NamedParameter) p).getName();
				} else {
					numberOrName = String.valueOf(i + 1);
				}
				printTextElement(qname("td"), null, numberOrName);
				printTextElement(qname("td"), null, p.getType().name());
				printTableData(p.getValue());
				printEndElement();
			}
			printEndElement();

			printEndElement();
		}
	}

	private void printTableData(Object value) {

		if (value instanceof Array) {
			Array sqlArray = (Array) value;
			try {
				Object[] array = (Object[]) sqlArray.getArray();
				printStartElement(qname("td"));
				printArray(array);
				printEndElement();
			} catch (SQLException e) {
				log.log(Level.SEVERE, "Unable to format array", e);
				printTableData(String.valueOf(value));
			}
		} else {
			Map<QName, String> attributes = null;
			if (value instanceof Number) {
				attributes = singleAttribute(qname("class"), "number");
			} else if (value instanceof Boolean) {
				attributes = singleAttribute(qname("class"), "boolean");
			}
			printTextElement(qname("td"), attributes, String.valueOf(value));
		}
	}

	private void printArray(Object[] array) {
		printStartElement(qname("ul"));
		for (Object o : array) {
			if (o instanceof Object[]) {
				printStartElement(qname("li"));
				printTextElement(qname("p"), null, "nested array:");
				printArray((Object[]) o);
				printEndElement();
			} else {
				printTextElement(qname("li"), null, String.valueOf(o));
			}
		}
		printEndElement();
	}

	@Override
	public void writeStartResultSet(ColumnsHeader header) {
		super.writeStartResultSet(header);
		resultSetCounter++;
		printEmptyElement(qname("hr"), null);
		printTextElement(qname("h3"), null, "Result set #" + resultSetCounter);
		printStartElement(qname("table"));
		printStartElement(qname("thead"));
		printStartElement(qname("tr"));
		for (ColumnDescriptor cd : header.getColumnDescriptors()) {
			// TODO: type
			printTextElement(qname("td"), null, cd.getLabel());
		}
		printEndElement();
		printEndElement();

		printStartElement(qname("tbody"));
	}

	@Override
	public void writeEndResultSet() {
		super.writeEndResultSet();
		printEndElement();
		printEndElement();
		printTextElement(qname("p"), null, "Record count: " + getCurrentRowCount());
	}

	@Override
	public void writeStartRow() {
		super.writeStartRow();
		printStartElement(qname("tr"));
	}

	@Override
	public void writeColumnValue(Object value) {
		super.writeColumnValue(value);
		printTableData(value);
	}

	@Override
	public void writeEndRow() {
		super.writeEndRow();
		printEndElement();
	}

	@Override
	public void writeStartStatement() {
		super.writeStartStatement();
		statementCounter++;
		printEmptyElement(qname("hr"), null);
		printTextElement(qname("h2"), null, "SQL statement #" + statementCounter);
		resultSetCounter = 0;
		updatesResultCounter = 0;
	}

	@Override
	public void writeUpdatesResult(int updatedRowsCount) {
		super.writeUpdatesResult(updatedRowsCount);
		updatesResultCounter++;
		printEmptyElement(qname("hr"), null);
		printTextElement(qname("h3"), null, "Updates result #" + updatesResultCounter);
		printTextElement(qname("p"), null, "Updated rows: " + updatedRowsCount);
	}
}
