/**
 * SQL-DK
 * Copyright © 2013 František Kučera (frantovo.cz)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package info.globalcode.sql.dk;

import info.globalcode.sql.dk.CLIParser.Tokens;
import static info.globalcode.sql.dk.CLIParser.TYPE_NAME_SEPARATOR;
import info.globalcode.sql.dk.InfoLister.InfoType;
import java.io.ByteArrayInputStream;
import java.util.Collection;
import static org.testng.Assert.*;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

/**
 *
 * @author Ing. František Kučera (frantovo.cz)
 */
public class CLIParserTest {

	private static final String DATABASE_NAME_1 = "some database 1";
	private static final String SQL_1 = "SELECT * FROM table1";
	private static final String DATA_1 = "aaa";
	private static final String DATA_2 = "bbb";
	private static final String DATA_3 = "ccc";
	private static final String NAME_1 = "param1";
	private static final String NAME_2 = "param2";
	private static final String NAME_3 = "param3";
	private CLIParser parser;

	@BeforeMethod
	public void setUpMethod() throws Exception {
		parser = new CLIParser();
	}

	private CLIOptions parseOptions(String[] args) throws CLIParserException {
		return parser.parseOptions(args, new ByteArrayInputStream("".getBytes()));
	}

	@Test
	public void testParseOptions_QueryNow_NoParams() throws InvalidOptionsException, CLIParserException {
		String[] args = new String[]{
			Tokens.DB, DATABASE_NAME_1,
			Tokens.SQL, SQL_1};
		CLIOptions options = parseOptions(args);
		options.validate();

		assertEquals(options.getDatabaseName(), DATABASE_NAME_1);
		assertEquals(options.getSql(), SQL_1);
		assertEquals(options.getMode(), CLIOptions.MODE.QUERY_NOW);
		assertTrue(options.getNamedParameters().isEmpty(), "Named parameters should be empty.");
		assertTrue(options.getNumberedParameters().isEmpty(), "Numbered parameters should be empty.");
	}

	@Test
	public void testParseOptions_QueryNow_Numbered() throws InvalidOptionsException, CLIParserException {
		String[] args = new String[]{
			Tokens.DB, DATABASE_NAME_1,
			Tokens.SQL, SQL_1,
			Tokens.DATA, DATA_1, DATA_2, DATA_3};
		CLIOptions options = parseOptions(args);
		options.validate();

		assertEquals(options.getDatabaseName(), DATABASE_NAME_1);
		assertEquals(options.getSql(), SQL_1);
		assertEquals(options.getMode(), CLIOptions.MODE.QUERY_NOW);
		assertEquals(options.getNumberedParameters().size(), 3);
		assertEquals(options.getNumberedParameters().get(0).getValue(), DATA_1);
		assertEquals(options.getNumberedParameters().get(1).getValue(), DATA_2);
		assertEquals(options.getNumberedParameters().get(2).getValue(), DATA_3);
		assertEquals(options.getNumberedParameters().get(0).getType(), Parameter.DEFAULT_TYPE);
		assertEquals(options.getNumberedParameters().get(1).getType(), Parameter.DEFAULT_TYPE);
		assertEquals(options.getNumberedParameters().get(2).getType(), Parameter.DEFAULT_TYPE);
	}

	@Test
	public void testParseOptions_QueryNow_Numbered_withTypes() throws InvalidOptionsException, CLIParserException {
		String[] args = new String[]{
			Tokens.DB, DATABASE_NAME_1,
			Tokens.SQL, SQL_1,
			Tokens.TYPES, " INTEGER,VARCHAR, BOOLEAN",
			Tokens.DATA, DATA_1, DATA_2, DATA_3};
		CLIOptions options = parseOptions(args);
		options.validate();

		assertEquals(options.getDatabaseName(), DATABASE_NAME_1);
		assertEquals(options.getSql(), SQL_1);
		assertEquals(options.getMode(), CLIOptions.MODE.QUERY_NOW);
		assertEquals(options.getNumberedParameters().size(), 3);
		assertEquals(options.getNumberedParameters().get(0).getValue(), DATA_1);
		assertEquals(options.getNumberedParameters().get(1).getValue(), DATA_2);
		assertEquals(options.getNumberedParameters().get(2).getValue(), DATA_3);
		assertEquals(options.getNumberedParameters().get(0).getType(), SQLType.INTEGER);
		assertEquals(options.getNumberedParameters().get(1).getType(), SQLType.VARCHAR);
		assertEquals(options.getNumberedParameters().get(2).getType(), SQLType.BOOLEAN);
	}

	@Test
	public void testParseOptions_QueryNow_Named() throws InvalidOptionsException, CLIParserException {
		String[] args = new String[]{
			Tokens.DB, DATABASE_NAME_1,
			Tokens.SQL, SQL_1,
			Tokens.DATA_NAMED, NAME_1, DATA_1, NAME_2, DATA_2, NAME_3, DATA_3};
		CLIOptions options = parseOptions(args);
		options.validate();

		assertEquals(options.getDatabaseName(), DATABASE_NAME_1);
		assertEquals(options.getSql(), SQL_1);
		assertEquals(options.getMode(), CLIOptions.MODE.QUERY_NOW);
		assertEquals(options.getNamedParameters().size(), 3);
		assertNamedParameter(options.getNamedParameters(), NAME_1, DATA_1, Parameter.DEFAULT_TYPE);
		assertNamedParameter(options.getNamedParameters(), NAME_2, DATA_2, Parameter.DEFAULT_TYPE);
		assertNamedParameter(options.getNamedParameters(), NAME_3, DATA_3, Parameter.DEFAULT_TYPE);
	}

	@Test
	public void testParseOptions_QueryNow_Named_withTypes() throws InvalidOptionsException, CLIParserException {
		String[] args = new String[]{
			Tokens.DB, DATABASE_NAME_1,
			Tokens.SQL, SQL_1,
			Tokens.NAME_PREFIX, "$",
			Tokens.TYPES, " " + NAME_1 + TYPE_NAME_SEPARATOR + "INTEGER" + "," + NAME_3 + TYPE_NAME_SEPARATOR + "BOOLEAN",
			Tokens.DATA_NAMED, NAME_1, DATA_1, NAME_2, DATA_2, NAME_3, DATA_3};
		CLIOptions options = parseOptions(args);
		options.validate();

		assertEquals(options.getDatabaseName(), DATABASE_NAME_1);
		assertEquals(options.getSql(), SQL_1);
		assertEquals(options.getMode(), CLIOptions.MODE.QUERY_NOW);
		assertEquals(options.getNamedParameters().size(), 3);
		assertNamedParameter(options.getNamedParameters(), NAME_1, DATA_1, SQLType.INTEGER);
		assertNamedParameter(options.getNamedParameters(), NAME_2, DATA_2, Parameter.DEFAULT_TYPE);
		assertNamedParameter(options.getNamedParameters(), NAME_3, DATA_3, SQLType.BOOLEAN);
	}

	private void assertNamedParameter(Collection<NamedParameter> params, String name, Object value, SQLType type) {
		for (NamedParameter p : params) {
			if (name.equals(p.getName())) {
				assertEquals(p.getValue(), value, "value does not match – name: " + name);
				assertEquals(p.getType(), type, "value does not match – name: " + name);
				return;
			}
		}
		fail("Named parameter not found: " + name);
	}

	@Test
	public void testParseOptions_PrepareBatch() throws InvalidOptionsException, CLIParserException {
		String[] args = new String[]{
			Tokens.BATCH,
			Tokens.SQL, SQL_1};
		CLIOptions options = parseOptions(args);
		options.validate();

		assertEquals(options.getSql(), SQL_1);
		assertEquals(options.getMode(), CLIOptions.MODE.PREPARE_BATCH);
	}

	@Test
	public void testParseOptions_ExecuteBatch() throws InvalidOptionsException, CLIParserException {
		String[] args = new String[]{
			Tokens.BATCH,
			Tokens.DB, DATABASE_NAME_1};
		CLIOptions options = parseOptions(args);
		options.validate();

		assertEquals(options.getDatabaseName(), DATABASE_NAME_1);
		assertEquals(options.getMode(), CLIOptions.MODE.EXECUTE_BATCH);
	}

	@Test
	public void testParseOptions_ShowInfo_Help() throws InvalidOptionsException, CLIParserException {
		String[] args = new String[]{Tokens.INFO_HELP};
		CLIOptions options = parseOptions(args);
		options.validate();

		assertEquals(options.getMode(), CLIOptions.MODE.JUST_SHOW_INFO);
		assertEquals(options.getShowInfo().size(), 1);
		assertTrue(options.getShowInfo().contains(InfoType.HELP));
	}
}